#ifndef MACHINE_HPP
#define MACHINE_HPP

#include <vector>
#include "data.hpp"

/// A virtual machine that executes programs while maintaining its state
class machine
{
	public:
	struct arg_value;

    // A function/instruction_index pair
	struct address {
		std::list<function>::const_iterator fun;
		std::list<stem>::const_iterator ins;
	};

	private:
	// A frame of execution
	struct frame {
		// address of the next instruction to execute (instruct pointer)
		address ip;

		// list of arguments passed to the active call
		std::list<arg_value> args;
	};

	// The currently loaded program
	std::list<function> m_pgm;

	// The current frame stack (current exec frame is at back())
	std::vector<frame> m_exec_stack;

	// Pushes a new exec frame on the stack while checking for stack
	// overflow
	void push_frame(frame const& f);

	// Pops the top exec frame out!
	// Returns false if the stack is empty after the operation
	bool pop_frame()
	{ m_exec_stack.pop_back(); if(m_exec_stack.empty()) return false; else return true; }

	// Look in the current frame for the value of an argument and return it
	// if not found return end()
	std::list<arg_value>::const_iterator fetch_arg_value(char name) const;

	// Look in the program for a function given its name (returns end() if not found)
	std::list<function>::const_iterator fetch_function(char name) const;

	// current exec frame
	frame& cur_frame() { return m_exec_stack.back(); }
	frame const& cur_frame() const { return m_exec_stack.back(); }

	// evaluate an expression (1+3-D-R+P-4...
	int eval(exp const& e) const;

	public:

	// An argument value passed to the called function (either an integer or an
	// instruction string like 'ssrssrsll')
	struct arg_value {
		char name;
		enum { number_type, ins_type } type;
		int number;
		std::vector<char> ins;
	};

    // Access to th loaded program.
    std::list<function> const& program() const { return m_pgm; }

	// Loads a program and prepares to execute it
	void load(std::list<function> const& pgm, std::list<arg_value> const& arg=std::list<arg_value>());

	// Executes some instructions and appends the results in res.
	// if there is nothing more to execute (i.e program finished)
	// returns false otherwise returns true;
	bool exec(std::vector<char>& res);

	// Instruction pointer
	address const& ip() const { return cur_frame().ip; }
};


#endif
